"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.QueryMetadata = void 0;
exports.query = query;
const core_1 = require("@loopback/core");
const repository_1 = require("@loopback/repository");
const assert_1 = require("tily/assert");
const query_1 = require("../query");
/**
 * Metadata for a query
 */
class QueryMetadata {
    /**
     * Constructor for QueryMetadata
     *
     * @param modelOrRepo - Name or class of the model. If the value is a string and
     * `dataSource` is not present, it will be treated as the name of a predefined
     * select query
     * @param dataSource - Name or instance of the data source
     *
     * For example:
     *
     * - new QueryMetadata(repoName);
     * - new QueryMetadata(modelName, dataSourceName);
     * - new QueryMetadata(modelClass, dataSourceInstance);
     * - new QueryMetadata(modelName, dataSourceInstance);
     * - new QueryMetadata(modelClass, dataSourceName);
     */
    constructor(modelOrRepo, dataSource) {
        this.name = typeof modelOrRepo === 'string' && dataSource === undefined ? modelOrRepo : undefined;
        this.modelName = typeof modelOrRepo === 'string' && dataSource != null ? modelOrRepo : undefined;
        this.modelClass = typeof modelOrRepo === 'function' ? modelOrRepo : undefined;
        this.dataSourceName = typeof dataSource === 'string' ? dataSource : undefined;
        this.dataSource = typeof dataSource === 'object' ? dataSource : undefined;
    }
}
exports.QueryMetadata = QueryMetadata;
function query(modelOrRepo, dataSource) {
    // if string, repository or not a model ctor,
    // keep it a string / assign to ctor's name (string) for DI
    const stringOrModel = typeof modelOrRepo !== 'string' && !modelOrRepo.prototype.getId ? modelOrRepo.name : modelOrRepo;
    const meta = new QueryMetadata(stringOrModel, dataSource);
    return function (target, key, 
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    descriptorOrIndex) {
        if (key || typeof descriptorOrIndex === 'number') {
            // Use repository or query-factory to create a query from model + dataSource
            (0, core_1.inject)('', meta, resolve)(target, key, descriptorOrIndex);
            return;
        }
        // Mixin repository into the class
        throw new Error('Class level @query is not implemented');
    };
}
/**
 * Resolve the @query injection
 * @param ctx - Context
 * @param injection - Injection metadata
 */
async function resolve(ctx, injection) {
    const meta = injection.metadata;
    if (meta.name) {
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const repo = (await ctx.get('repositories.' + meta.name));
        if ((0, query_1.isQuery)(repo.query)) {
            return repo.query;
        }
        return new query_1.DefaultQuery(repo);
    }
    let modelClass = meta.modelClass;
    if (meta.modelName) {
        modelClass = (await ctx.get('models.' + meta.modelName));
    }
    if (!modelClass) {
        throw new Error('Invalid query config: neither modelClass nor modelName was specified.');
    }
    let dataSource = meta.dataSource;
    if (meta.dataSourceName) {
        dataSource = await ctx.get('datasources.' + meta.dataSourceName);
    }
    (0, assert_1.assert)(dataSource instanceof repository_1.juggler.DataSource, 'DataSource must be provided');
    return new query_1.DefaultQuery(modelClass, dataSource);
}
//# sourceMappingURL=query.decorator.js.map