"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JoinResolver = void 0;
exports.parseRelationChain = parseRelationChain;
const tslib_1 = require("tslib");
const repository_1 = require("@loopback/repository");
const debug_1 = tslib_1.__importDefault(require("debug"));
const includes_1 = tslib_1.__importDefault(require("tily/array/includes"));
const toArray_1 = tslib_1.__importDefault(require("tily/array/toArray"));
const assert_1 = require("tily/assert");
const array_1 = tslib_1.__importDefault(require("tily/is/array"));
const plainObject_1 = tslib_1.__importDefault(require("tily/is/plainObject"));
const string_1 = require("tily/is/string");
const relation_1 = require("../../../relation");
const session_1 = require("../../../session");
const clause_1 = require("../clause");
const types_1 = require("../types");
const debug = (0, debug_1.default)('bleco:query:join');
class JoinResolver extends clause_1.ClauseResolver {
    resolve(qb, filter, session = session_1.QuerySession.create()) {
        debug(`Resolving where clause for model ${this.entityClass.modelName}:`, filter, session);
        const { where, order } = filter;
        if (!where && !order) {
            debug('No where or order found, skip resolving');
            return;
        }
        if ((typeof where !== 'object' || Array.isArray(where)) && !order) {
            debug('Invalid where: %j', where);
            return;
        }
        if (where) {
            for (const key of extractKeys(where)) {
                this.compile(key, session, session.relationWhere);
            }
        }
        if (order) {
            for (const key of (0, toArray_1.default)(order)) {
                this.compile(retrieveOrderKey(key), session, session.relationOrder);
            }
        }
        this.buildJoins(qb, session);
    }
    buildJoins(qb, session) {
        // TODO: support MySQL and Postgres "uuid" type for id column: https://github.com/Wikodit/loopback-connector-postgresql/blob/develop/lib/postgresql.js#L965
        const orm = this.orm;
        for (const relationJoin of session.relationJoins) {
            (0, assert_1.assert)(relationJoin.relation.keyFrom, 'relation.keyFrom is required');
            (0, assert_1.assert)(relationJoin.relation.keyTo, 'relation.keyTo is required');
            const targetModel = relationJoin.model;
            if (!orm.getModelDefinition(targetModel)) {
                throw new Error(`Model "${relationJoin.model}" is not defined in the current datasource for relation "${relationJoin.relationPath}".`);
            }
            qb.leftJoin({
                [orm.escapeName(relationJoin.prefix + orm.table(targetModel))]: orm.tableEscaped(targetModel),
            }, 
            // eslint-disable-next-line @typescript-eslint/no-shadow
            qb => {
                var _a;
                qb.on(orm.columnEscaped(relationJoin.parentModel, relationJoin.relation.keyFrom, true, relationJoin.parentPrefix), orm.columnEscaped(targetModel, relationJoin.relation.keyTo, true, relationJoin.prefix));
                if ((_a = relationJoin.polymorphic) === null || _a === void 0 ? void 0 : _a.value) {
                    qb.onVal(orm.columnEscaped(relationJoin.parentModel, relationJoin.polymorphic.discriminator, true, relationJoin.parentPrefix), relationJoin.polymorphic.value);
                }
            });
        }
    }
    compile(key, session, constraints) {
        var _a, _b;
        const { definition } = this.entityClass;
        const parsed = parseRelationChain(definition, key);
        if (!parsed) {
            return;
        }
        const { relationChain, property, propertyKey } = parsed;
        const relationPath = this.entityClass.modelName;
        let parentPrefix = '';
        let parentEntity = this.entityClass;
        let join;
        let potentialRelationPath = relationPath;
        for (let i = 0; i < relationChain.length; i++) {
            const modelDefinition = parentEntity.definition;
            // Build a prefix for alias to prevent conflict
            const prefix = nextPrefix(session, i);
            const { name: candidateRelation, polymorphicValue } = relationChain[i];
            potentialRelationPath += '.' + candidateRelation;
            if (!modelDefinition) {
                debug('No definition for model %s', parentEntity);
                break;
            }
            if (!(candidateRelation in modelDefinition.relations)) {
                debug('No relation for model %s', parentEntity.modelName);
                break;
            }
            if ((0, includes_1.default)(candidateRelation, (_a = definition.settings.hiddenRelations) !== null && _a !== void 0 ? _a : [])) {
                debug('Hidden relation for model %s skipping', parentEntity.modelName);
                break;
            }
            let relation = modelDefinition.relations[candidateRelation];
            // Only supports belongsTo, hasOne and hasMany
            if (!relation_1.SupportedRelationTypes.includes(relation.type)) {
                debug('Invalid relation type for model %s for left join', parentEntity.modelName);
                break;
            }
            relation = (0, relation_1.resolveRelation)(relation);
            const target = relation.target();
            let polymorphic;
            // Check if the relation is already joined
            join = session.relationJoins.find(j => potentialRelationPath === j.relationPath);
            if (!join) {
                // Join the relation
                if (relation.type === repository_1.RelationType.hasMany && relation.through) {
                    const throughEntity = relation.through.model();
                    polymorphic = (0, plainObject_1.default)(relation.through.polymorphic)
                        ? {
                            discriminator: relation.through.polymorphic.discriminator,
                            value: polymorphicValue,
                        }
                        : undefined;
                    // Add through relation join
                    session.addRelationJoin({
                        relationPath: `${relationPath}.-.${candidateRelation}`,
                        prefix,
                        parentPrefix,
                        parentModel: parentEntity.modelName,
                        relation: {
                            keyFrom: relation.keyFrom,
                            keyTo: relation.through.keyFrom,
                            name: `_${candidateRelation}_`,
                        },
                        model: throughEntity.modelName,
                    });
                    parentPrefix = prefix;
                    parentEntity = throughEntity;
                    relation = {
                        keyFrom: relation.through.keyTo,
                        keyTo: relation.keyTo,
                    };
                }
                else {
                    polymorphic =
                        'polymorphic' in relation && (0, plainObject_1.default)(relation.polymorphic)
                            ? {
                                discriminator: relation.polymorphic.discriminator,
                                value: polymorphicValue,
                            }
                            : undefined;
                }
                // add relation join
                join = session.addRelationJoin({
                    relationPath: potentialRelationPath,
                    prefix,
                    parentPrefix,
                    parentModel: parentEntity.modelName,
                    relation: {
                        ...relation,
                        name: candidateRelation,
                    },
                    model: (_b = polymorphic === null || polymorphic === void 0 ? void 0 : polymorphic.value) !== null && _b !== void 0 ? _b : target.modelName,
                    polymorphic,
                });
            }
            // Keep the prefix of the found join
            parentPrefix = join.prefix;
            // Keep the parentEntity for recursive JOIN
            parentEntity = target;
        }
        if (!join)
            return;
        // Keep what needed to build the WHERE or ORDER statement properly
        constraints[key] = {
            prefix: join.prefix,
            model: join.model,
            property: {
                ...property,
                key: propertyKey,
            },
        };
    }
}
exports.JoinResolver = JoinResolver;
function parseRelationChain(definition, key) {
    var _a, _b, _c, _d;
    const members = parseKey(key);
    let i = 0;
    let relation = definition.relations[(_a = members[i]) === null || _a === void 0 ? void 0 : _a.name];
    if (relation) {
        let next = relation.target().definition.relations[(_b = members[++i]) === null || _b === void 0 ? void 0 : _b.name];
        while (next) {
            relation = next;
            next = relation.target().definition.relations[(_c = members[++i]) === null || _c === void 0 ? void 0 : _c.name];
        }
    }
    // no relation found
    if (!relation) {
        if (!definition.properties[(_d = members[0]) === null || _d === void 0 ? void 0 : _d.name]) {
            throw new Error(`No relation and property found for key "${key}" in model "${definition.name}"`);
        }
        // ignore nested properties
        return;
    }
    // no property left
    if (i >= members.length) {
        return {
            relationChain: members,
        };
    }
    const relationChain = members.slice(0, i);
    const propertyName = members[i].name;
    const target = relation.target().definition;
    // property may be undefined if the relation is polymorphic and target is ancestor class
    const property = target.properties[propertyName];
    if (!hasPolymorphic(relation) && !property) {
        throw new Error(`"${propertyName}" is not in model "${target.name}" with relation chain "${relationChain
            .map(r => r.name)
            .join('.')}"`);
    }
    for (let j = i; j < members.length; j++) {
        if (members[j].polymorphicValue) {
            throw new Error(`polymorphic should not be with property: ${propertyName}`);
        }
    }
    return {
        relationChain,
        property,
        propertyKey: members
            .slice(i)
            .map(m => m.name)
            .join('.'),
    };
}
function parseKey(key) {
    return key.split('.').map(part => {
        const from = part.indexOf('(');
        const to = part.lastIndexOf(')');
        if ((from < 0 && to >= 0) || (from >= 0 && to < 0)) {
            throw new Error(`Invalid key "${key}"`);
        }
        if (from < 0) {
            return { name: part };
        }
        else {
            return { name: part.substring(0, from), polymorphicValue: part.substring(from + 1, to) };
        }
    });
}
function extractKeys(where, keys = new Set()) {
    for (const key in where) {
        if (types_1.GroupOperators.includes(key)) {
            // {and|or: [{...}]}
            const clauses = where[key];
            if (Array.isArray(clauses)) {
                for (const clause of clauses) {
                    extractKeys(clause, keys);
                }
            }
        }
        else if (key === types_1.Directives.EXPR) {
            // extract $expr
            const expr = where.$expr;
            const op = Object.keys(expr)[0];
            if (!op)
                continue;
            const value = expr[op];
            (0, assert_1.assert)((0, array_1.default)(value) && value.length === 2, `$expr->${op} must be an array value with 2 elements`);
            for (const v of value) {
                if ((0, string_1.isString)(v) && v.startsWith('$')) {
                    keys.add(v.substring(1));
                }
            }
        }
        else if (key === types_1.Directives.JOIN) {
            const joins = where.$join;
            (0, toArray_1.default)(joins).forEach(j => keys.add(j));
        }
        else {
            keys.add(key);
        }
    }
    return keys;
}
function hasPolymorphic(relation) {
    const rel = 'through' in relation && relation.through ? relation.through : relation;
    return 'polymorphic' in rel && (0, plainObject_1.default)(rel.polymorphic);
}
function retrieveOrderKey(key) {
    const t = key.split(/[\s,]+/);
    return t.length === 1 ? key : t[0];
}
function nextPrefix(session, i) {
    return `t_${session.nextSeq()}_${i}_`;
}
//# sourceMappingURL=join.js.map