"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.WhereResolver = exports.DefaultOperatorHandlerRegistry = exports.OperatorHandlerRegistry = void 0;
const tslib_1 = require("tslib");
/* eslint-disable @typescript-eslint/no-floating-promises, @typescript-eslint/no-misused-promises, @typescript-eslint/no-explicit-any, @typescript-eslint/no-shadow */
const filter_1 = require("@loopback/filter");
const debug_1 = tslib_1.__importDefault(require("debug"));
const assert_1 = require("tily/assert");
const array_1 = tslib_1.__importDefault(require("tily/is/array"));
const empty_1 = tslib_1.__importDefault(require("tily/is/empty"));
const object_1 = tslib_1.__importDefault(require("tily/is/object"));
const plainObject_1 = tslib_1.__importDefault(require("tily/is/plainObject"));
const string_1 = require("tily/is/string");
const each_1 = tslib_1.__importDefault(require("tily/object/each"));
const session_1 = require("../../../session");
const utils_1 = require("../../../utils");
const clause_1 = require("../clause");
const types_1 = require("../types");
const debug = (0, debug_1.default)('bleco:query:where');
class OperatorHandlerRegistry {
    constructor() {
        this.handlers = {};
        this.init();
    }
    init() {
        this.register('=', this.eq());
        this.register('!=', this.neq());
        this.register('<', this.comparison('<'));
        this.register('<=', this.comparison('<='));
        this.register('>', this.comparison('>'));
        this.register('>=', this.comparison('>='));
        this.register('!', this.not());
        this.register('eq', this.eq());
        this.register('neq', this.neq());
        this.register('lt', this.comparison('<'));
        this.register('lte', this.comparison('<='));
        this.register('gt', this.comparison('>'));
        this.register('gte', this.comparison('>='));
        this.register('not', this.not());
        this.register('in', this.whereForArrayValue('whereIn'));
        this.register('inq', this.whereForArrayValue('whereIn'));
        this.register('nin', this.whereForArrayValue('whereNotIn'));
        this.register('between', this.whereForArrayValue('whereBetween'));
        this.register('like', this.whereLike('whereLike'));
        this.register('nlike', this.comparison('not like'));
        this.register('ilike', this.whereLike('whereILike'));
        this.register('nilike', this.comparison('not ilike'));
        this.register('or', this.logical('orWhere'));
        this.register('and', this.logical('where'));
    }
    register(operator, handler) {
        this.handlers[operator] = handler;
    }
    get(operator) {
        const h = this.handlers[operator];
        if (!h) {
            throw new Error(`Operator "${operator}" is not supported`);
        }
        return h;
    }
    eq() {
        return (qb, condition) => {
            debug('- eq:', condition);
            const { key, value, params } = condition;
            if (condition.directive === '$expr') {
                if (debug.enabled)
                    debug('whereRaw:', `${key} = ${value}`, params);
                return qb.whereRaw(`${key} = ${value}`, params);
            }
            if (value == null) {
                if (debug.enabled)
                    debug('whereNull:', key);
                qb.whereNull(key);
            }
            else {
                if (debug.enabled)
                    debug('where:', key, value);
                qb.where(key, '=', value);
            }
        };
    }
    neq() {
        return (qb, condition) => {
            debug('- neq:', condition);
            const { key, value, params } = condition;
            if (condition.directive === '$expr') {
                if (debug.enabled)
                    debug('whereRaw:', `${key} != ${value}`, params);
                return qb.whereRaw(`${key} != ${value}`, params);
            }
            if (value == null) {
                if (debug.enabled)
                    debug('whereNotNull:', key);
                qb.whereNotNull(key);
            }
            else {
                if (debug.enabled)
                    debug('where:', key, value);
                qb.where(key, '!=', value);
            }
        };
    }
    comparison(operator) {
        return (qb, condition) => {
            debug('- comparison:', condition);
            const { key, value, params } = condition;
            if (condition.directive === '$expr') {
                if (debug.enabled)
                    debug('whereRaw:', `${key} ${operator} ${value}`, params);
                return qb.whereRaw(`${key} ${operator} ${value}`, params);
            }
            if (debug.enabled)
                debug('where:', key, value);
            qb.where(key, operator, value);
        };
    }
    not() {
        return function (qb, condition, session) {
            debug('- not:', condition);
            const { value } = condition;
            if (debug.enabled)
                debug('whereNot:', value);
            qb.whereNot(qb => this.build(qb, value, session));
        };
    }
    whereForArrayValue(op) {
        return function (qb, condition) {
            // eslint-disable-next-line prefer-const
            let { key, value } = condition;
            if ((0, array_1.default)(value)) {
                value = value.map(v => (v == null ? null : v));
            }
            if (op === 'whereIn' && (0, empty_1.default)(value)) {
                value = [null];
            }
            else if (op === 'whereBetween') {
                const v1 = value[0] === undefined ? null : value[0];
                const v2 = value[1] === undefined ? null : value[1];
                value = [v1, v2];
            }
            debug('- [whereForArrayValue] %s: %O', op, { key, value });
            qb[op](key, value);
        };
    }
    whereLike(op, ignoreEmptyArray) {
        return function (qb, condition) {
            debug('- [whereLike] %s: %O', op, condition);
            const { key, value } = condition;
            qb[op](key, value == null ? null : value);
        };
    }
    logical(op) {
        return function (qb, condition, session) {
            debug('- logical:', op, condition);
            const { value } = condition;
            if ((0, array_1.default)(value)) {
                if ((0, empty_1.default)(value)) {
                    debug('- logical(array): empty array');
                    return;
                }
                debug('- logical: resolving orWhere for array value');
                qb.where(qb => value.forEach(v => qb[op](qb => this.build(qb, v, session))));
                return;
            }
            if (!(0, object_1.default)(value)) {
                debug('- logical: value is not object');
                return;
            }
            if (op === 'where') {
                debug('- logical: resolving where');
                return this.build(qb, value, session);
            }
            debug('- logical: resolving orWhere for object value');
            qb.where(qb => {
                (0, each_1.default)((v, k) => qb[op](qb => this.build(qb, { [k]: v }, session)), value);
            });
        };
    }
}
exports.OperatorHandlerRegistry = OperatorHandlerRegistry;
exports.DefaultOperatorHandlerRegistry = new OperatorHandlerRegistry();
class WhereResolver extends clause_1.ClauseResolver {
    constructor(entityClass, orm, registry = exports.DefaultOperatorHandlerRegistry) {
        super(entityClass, orm);
        this.registry = registry;
    }
    resolve(qb, filter, session) {
        const where = (0, filter_1.isFilter)(filter) ? filter.where : filter;
        if (!where) {
            debug('No where clause, skip resolving');
            return;
        }
        debug(`Resolving where clause for model ${this.entityClass.modelName}:`, where, session);
        this.build(qb, where, session_1.QuerySession.create(session));
    }
    build(qb, where, session) {
        debug(`- build: building where clause for model ${this.entityClass.modelName}:`, where);
        (0, each_1.default)((v, k) => {
            const condition = parseCondition(k, v);
            if (!condition) {
                return;
            }
            debug('- build: parsed clause:', condition);
            this.invoke(qb, condition, session);
        }, (0, utils_1.compactWhere)(where));
    }
    invoke(qb, rawCondition, session) {
        // skip escape
        qb.queryContext({ skipEscape: true });
        const condition = this.resolveCondition(rawCondition, session);
        debug('invoke: resolved condition:', condition);
        if (!types_1.GroupOperators.includes(condition.op) && !condition.key) {
            // skip empty condition
            return;
        }
        this.registry.get(condition.op).call(this, qb, condition, session);
    }
    resolveCondition(condition, session) {
        const { directive } = condition;
        let { key, value } = condition;
        const params = [];
        if (directive === '$expr') {
            [key, value] = [key, value].map(v => {
                if ((0, string_1.isString)(v) && v.startsWith('$')) {
                    return this.resolveKey(v.substring(1), session);
                }
                params.push(v);
                return '?';
            });
        }
        else if (key) {
            if ((0, string_1.isString)(key)) {
                key = this.resolveKey(key, session);
            }
            else {
                params.push(key);
                key = '?';
            }
        }
        return { ...condition, key, value, params };
    }
    resolveKey(key, session) {
        const { relationWhere } = session;
        const { definition } = this.entityClass;
        const props = definition.properties;
        if (props) {
            let constraint;
            let p = props[key];
            // TODO we should ignore hidden property for `where`?
            // if (p && includes(key, definition.settings.hiddenProperties ?? [])) {
            //   debug('Hidden prop for model %s skipping', definition.name);
            //   return '';
            // }
            if (p == null && (0, utils_1.isNested)(key)) {
                // See if we are querying nested json
                p = props[key.split('.')[0]];
            }
            // It may be a relation where
            if (p == null) {
                constraint = relationWhere === null || relationWhere === void 0 ? void 0 : relationWhere[key];
                if (constraint && !constraint.property) {
                    debug('Ignore relation where with key "%s" for no property provided', key);
                    return '';
                }
                p = constraint === null || constraint === void 0 ? void 0 : constraint.property;
            }
            if (p == null) {
                // Unknown property, ignore it
                debug('Unknown key "%s" is skipped for model "%s"', key, this.entityClass.modelName);
                return '';
            }
            return constraint
                ? this.orm.columnEscaped(constraint.model, p.key, true, constraint.prefix)
                : this.orm.columnEscaped(this.entityClass.modelName, key, session.hasRelations());
        }
        return key;
    }
}
exports.WhereResolver = WhereResolver;
function parseCondition(key, expression) {
    // skip relation where
    if (key === types_1.Directives.JOIN) {
        return;
    }
    if (expression === null) {
        return { key, op: '=', value: expression, expression };
    }
    if (types_1.GroupOperators.includes(key)) {
        return { key: '', op: key, value: expression, expression };
    }
    if (key === types_1.Directives.EXPR) {
        (0, assert_1.assert)((0, plainObject_1.default)(expression), '$expr must be an object');
        const op = Object.keys(expression)[0];
        const value = expression[op];
        (0, assert_1.assert)((0, array_1.default)(value) && value.length === 2, `$expr->${op} must be an array value with 2 elements`);
        return { directive: key, key: value[0], op, value: value[1], expression };
    }
    if ((0, plainObject_1.default)(expression)) {
        const op = Object.keys(expression)[0];
        return { key, op, value: expression[op], expression };
    }
    return { key, op: '=', value: expression, expression };
}
//# sourceMappingURL=where.js.map