"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SqlDriver = void 0;
const driver_1 = require("../../driver");
const session_1 = require("../../session");
const knex_1 = require("./knex");
const statement_1 = require("./statement");
const debug = require('debug')('bleco:query:driver:sql');
class SqlDriver extends driver_1.Driver {
    async find(model, filter, options) {
        const resolver = this.resolvers.get(model);
        filter = filter !== null && filter !== void 0 ? filter : {};
        this.applySortPolicy(model, filter);
        const [qb] = this.buildSelect(model, filter);
        resolver.resolveColumns(qb, filter);
        const s = qb.toSQL().toNative();
        if (debug.enabled) {
            debug(`Find with SQL: %s`, s.sql);
            debug(`Parameters: %o`, s.bindings);
        }
        const rows = await this.orm.execute(s.sql, s.bindings, options);
        return rows.map(row => this.orm.fromRow(model.modelName, row));
    }
    async count(model, where, options) {
        var _a, _b;
        const [qb, session] = this.buildSelect(model, { where });
        const builder = session.hasRelationJoins() ? this.knex(qb) : qb;
        builder.count('*', { as: 'cnt' });
        const s = builder.toSQL().toNative();
        if (debug.enabled) {
            debug(`Count with SQL: %s`, s.sql);
            debug(`Parameters: %o`, s.bindings);
        }
        const data = await this.orm.execute(s.sql, s.bindings, options);
        if (debug.enabled) {
            debug(`Count result: %o`, data);
        }
        return { count: (_b = (_a = data[0]) === null || _a === void 0 ? void 0 : _a.cnt) !== null && _b !== void 0 ? _b : 0 };
    }
    init() {
        this.knex = (0, knex_1.createKnex)(this.dataSource, this.options);
        this.resolvers = new statement_1.StatementResolvers(this.orm);
    }
    buildSelect(model, filter) {
        const transformer = this.resolvers.get(model);
        const session = session_1.QuerySession.create();
        const qb = this.knex(this.orm.tableEscaped(model.modelName)).queryContext({ skipEscape: true });
        transformer.resolveJoin(qb, filter, session);
        transformer.resolveWhere(qb, filter, session);
        transformer.resolveOrder(qb, filter, session);
        this.resolveLimit(qb, filter);
        if (session.hasRelationJoins()) {
            // groupBy ids to avoid duplication for joins
            qb.groupBy(this.orm.idNames(model.modelName).map(id => this.orm.columnEscaped(model.modelName, id, true)));
        }
        return [qb, session];
    }
    resolveLimit(qb, filter) {
        var _a;
        let limit = filter.limit;
        let offset = (_a = filter.skip) !== null && _a !== void 0 ? _a : filter.offset;
        if (isNaN(limit !== null && limit !== void 0 ? limit : 0)) {
            limit = 0;
        }
        if (isNaN(offset !== null && offset !== void 0 ? offset : 0)) {
            offset = 0;
        }
        if (!limit && !offset) {
            return;
        }
        if (limit) {
            qb.limit(limit);
        }
        if (offset) {
            qb.offset(offset);
        }
    }
}
exports.SqlDriver = SqlDriver;
//# sourceMappingURL=sql.js.map